import { strict as assert } from 'assert';
import express, { Request, Response, NextFunction } from 'express';
import axios from 'axios';
import fs from 'fs';
import path from 'path';
import https from 'https';
import { feathers, HookContext, Id, Application } from '@feathersjs/feathers';

import * as expressify from '../src';
import { RequestListener } from 'http';

describe('@feathersjs/express', () => {
  const service = {
    async get (id: Id) {
      return { id };
    }
  };

  it('exports .default, .original .rest, .notFound and .errorHandler', () => {
    assert.strictEqual(expressify.original, express);
    assert.strictEqual(typeof expressify.rest, 'function');
    assert.ok(expressify.notFound);
    assert.ok(expressify.errorHandler);
  });

  it('returns an Express application', () => {
    const app: Application = expressify.default(feathers());

    assert.strictEqual(typeof app, 'function');
  });

  it('allows to use an existing Express instance', () => {
    const expressApp = express();
    const app = expressify.default(feathers(), expressApp);

    assert.strictEqual(app, expressApp);
  });

  it('exports `express.rest`', () => {
    assert.ok(typeof expressify.rest === 'function');
  });

  it('returns a plain express app when no app is provided', () => {
    const app = expressify.default();

    assert.strictEqual(typeof app.use, 'function');
    assert.strictEqual(typeof app.service, 'undefined');
    assert.strictEqual(typeof app.services, 'undefined');
  });

  it('errors when app with wrong version is provided', () => {
    try {
      // @ts-ignore
      expressify.default({});
    } catch (e: any) {
      assert.strictEqual(e.message, '@feathersjs/express requires a valid Feathers application instance');
    }

    try {
      const app = feathers();
      app.version = '2.9.9';

      expressify.default(app);
    } catch (e: any) {
      assert.strictEqual(e.message, '@feathersjs/express requires an instance of a Feathers application version 3.x or later (got 2.9.9)');
    }

    try {
      const app = feathers();
      delete app.version;

      expressify.default(app);
    } catch (e: any) {
      assert.strictEqual(e.message, '@feathersjs/express requires an instance of a Feathers application version 3.x or later (got unknown)');
    }
  });

  it('Can use Express sub-apps', () => {
    const typedApp = feathers<{}>();
    const app = expressify.default(typedApp);
    const child = express();

    app.use('/path', child);
    assert.strictEqual((child as any).parent, app);
  });

  it('Can use express.static', () => {
    const app = expressify.default(feathers());

    app.use('/path', expressify.static(__dirname));
  });

  it('has Feathers functionality', async () => {
    const app = expressify.default(feathers());

    app.use('/myservice', service);

    app.hooks({
      after: {
        get (hook: HookContext) {
          hook.result.fromAppHook = true;
        }
      }
    });

    app.service('myservice').hooks({
      after: {
        get (hook: HookContext) {
          hook.result.fromHook = true;
        }
      }
    });

    const data = await app.service('myservice').get(10);

    assert.deepStrictEqual(data, {
      id: 10,
      fromHook: true,
      fromAppHook: true
    });
  });

  it('can register a service and start an Express server', async () => {
    const app = expressify.default(feathers());
    const response = {
      message: 'Hello world'
    };

    app.use('/myservice', service);
    app.use((_req: Request, res: Response) => res.json(response));

    const server = await app.listen(8787);
    const data = await app.service('myservice').get(10);

    assert.deepStrictEqual(data, { id: 10 });

    const res = await axios.get<any>('http://localhost:8787');
    assert.deepStrictEqual(res.data, response);

    await new Promise(resolve => server.close(() => resolve(server)));
  });

  it('.listen calls .setup', async () => {
    const app = expressify.default(feathers());
    let called = false;

    app.use('/myservice', {
      async get (id: Id) {
        return { id };
      },

      async setup (appParam, path) {
        assert.strictEqual(appParam, app);
        assert.strictEqual(path, 'myservice');
        called = true;
      }
    });

    const server = await app.listen(8787);

    assert.ok(called);
    await new Promise(resolve => server.close(() => resolve(server)));
  });

  it('passes middleware as options', () => {
    const feathersApp = feathers();
    const app = expressify.default(feathersApp);
    const oldUse = feathersApp.use;
    const a = (_req: Request, _res: Response, next: NextFunction) => next();
    const b = (_req: Request, _res: Response, next: NextFunction) => next();
    const c = (_req: Request, _res: Response, next: NextFunction) => next();
    const service = {
      async get (id: Id) {
        return { id };
      }
    };

    feathersApp.use = function (path, serviceArg, options) {
      assert.strictEqual(path, '/myservice');
      assert.strictEqual(serviceArg, service);
      assert.deepStrictEqual(options.express, {
        before: [a, b],
        after: [c]
      });
      return (oldUse as any).apply(this, arguments);
    };

    app.use('/myservice', a, b, service, c);
  });

  it('Express wrapped and context.app are the same', async () => {
    const app = expressify.default(feathers());

    app.use('/test', {
      async get (id: Id) {
        return { id };
      }
    });

    app.service('test').hooks({
      before: {
        get: [context => {
          assert.ok(context.app === app);
        }]
      }
    });

    assert.deepStrictEqual(await app.service('test').get('testing'), {
      id: 'testing'
    });
  });

  it('Works with HTTPS', done => {
    const todoService = {
      async get (name: Id) {
        return {
          id: name,
          description: `You have to do ${name}!`
        };
      }
    };

    const app = expressify.default(feathers()).configure(expressify.rest());

    app.use('/secureTodos', todoService);

    const httpsServer = https.createServer({
      key: fs.readFileSync(path.join(__dirname, '..', '..', 'tests', 'resources', 'privatekey.pem')),
      cert: fs.readFileSync(path.join(__dirname, '..', '..', 'tests', 'resources', 'certificate.pem')),
      rejectUnauthorized: false,
      requestCert: false
    }, app as unknown as RequestListener).listen(7889);

    app.setup(httpsServer);

    httpsServer.on('listening', function () {
      const instance = axios.create({
        httpsAgent: new https.Agent({
          rejectUnauthorized: false
        })
      });

      instance.get<any>('https://localhost:7889/secureTodos/dishes').then(response => {
        assert.ok(response.status === 200, 'Got OK status code');
        assert.strictEqual(response.data.description, 'You have to do dishes!');
        httpsServer.close(() => done());
      }).catch(done);
    });
  });
});
